#include <cstring>
#include <string>
#include <vector>
#include <cstdio>
#include <csignal>
#include <list>

#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>

using std::string;
using std::vector;
using std::list;
using std::printf;
using std::sprintf;
using std::strcat;
using std::fprintf;
using std::fflush;
using std::signal;

/* Nul FD class */
static class nul_fd
{
    int fd;
public:
    nul_fd() : fd(open("/dev/null", O_WRONLY)) { }
    ~nul_fd() { if(fd>=0)close(fd); }
    operator int () { return fd; }
} nul_fd;

static enum
{
    make_rar,
    make_gzip,
    make_bzip2
} op_mode = make_rar;

/* Does not return */
static void build_arch(const string &dest, const list<string> &fn)
{
    switch(op_mode)
    {
        case make_rar:
        {
            vector<const char *> argv(10 + fn.size());
            argv[0] = "rar";
            argv[1] = "a";
            argv[2] = dest.c_str();
            argv[3] = "-mm";
            argv[4] = "-m5";
            argv[5] = "-s";
            argv[6] = "-ol";
            argv[7] = "-ds";
            argv[8] = "--";
            argv[9 + fn.size()] = NULL;
            
            unsigned a = 9;
            list<string>::const_iterator i;
            for(i = fn.begin(); i != fn.end(); ++i)
                argv[a++] = i->c_str();
            
            execvp(argv[0], const_cast<char *const *> (&argv[0]));
            break;
        }
        case make_gzip:
        case make_bzip2:
        {
            int pip[2];
            pipe(pip);
            
            int p = fork();
            if(!p)
            {
                dup2(pip[1], 1);
                close(pip[0]);
                close(pip[1]);
                
                vector<const char *> argv(5 + fn.size());
                argv[0] = "tar";
                argv[1] = "cvvf";
                argv[2] = "-";
                argv[3 + fn.size()] = NULL;
                
                unsigned a = 3;
                list<string>::const_iterator i;
                for(i = fn.begin(); i != fn.end(); ++i)
                    argv[a++] = i->c_str();
                
                execvp(argv[0], const_cast<char *const *> (&argv[0]));
            }
#if 0
            p = fork();
            if(!p)
#endif
            {
                 dup2(pip[0], 0);
                 close(pip[0]);
                 close(pip[1]);
                 
                 int f = open(dest.c_str(), O_WRONLY | O_CREAT | O_TRUNC);
                 if(f < 0)
                 {
                     perror(dest.c_str());
                     _exit(0);
                 }
                 dup2(f, 1);
                 close(f);
                 
                 const char *prog = "gzip";
                 if(op_mode == make_bzip2)prog = "bzip2";
                 
                 execlp(prog, prog, "-c9", NULL);
            }
#if 0
            close(pip[0]);
            close(pip[1]);
            wait(NULL);
            _exit(0);
#endif
        }
    }
}

class similarity_tester
{
    static const string tempfile(int p=0)
    {
        char Buf[128];
        sprintf(Buf, "raropt__%d", p?p:getpid());
        switch(op_mode)
        {
            case make_rar: strcat(Buf, ".rar"); break;
            case make_gzip: strcat(Buf, ".tar.gz"); break;
            case make_bzip2: strcat(Buf, ".tar.bz2"); break;
        }
        return Buf;
    };

    class rec { public: int pid; unsigned total; unsigned key;
                rec(int p, unsigned t, unsigned k)
                : pid(p), total(t), key(k) { } };
    vector<rec> tasks;
    int laststatus;
    unsigned processlimit;

    double finish(unsigned which)
    {
        const string tmpfn = tempfile(tasks[which].pid);
        struct stat st;
        
        stat(tmpfn.c_str(), &st);
        unlink(tmpfn.c_str());
        
        double result = tasks[which].total / (double)st.st_size;
        
        tasks.erase(tasks.begin()+which);
        
        return result;
    }

public:
    similarity_tester(unsigned maxprocess=4) : laststatus(0), processlimit(maxprocess)
    {
    }
    
    void die()
    {
        for(unsigned a=0; a<tasks.size(); ++a)
        {
        	fprintf(stderr, "Sending SIGTERM to process %d\n", tasks[a].pid);
        	kill(tasks[a].pid, SIGTERM);
        }        
        
        while(wait(NULL) >= 0);
        
        for(unsigned a=0; a<tasks.size(); ++a)
        {
        	const string tmpfn = tempfile(tasks[a].pid);
        	fprintf(stderr, "Erasing %s\n", tmpfn.c_str());
        	unlink(tmpfn.c_str());
        }
        tasks.clear();
    }
    
    void set_process_limit(unsigned n) { processlimit = n; }
    
    void add(unsigned key, const list<string> &fn)
    {
        int p = fork();
        if(!p)
        {
            dup2(nul_fd, 1);
            dup2(nul_fd, 2);
            
            build_arch(tempfile(), fn);
            _exit(0);
        }
        if(p < 0)
            return;

        struct stat st;
        unsigned total = 0;
        
        list<string>::const_iterator i;
        for(i = fn.begin(); i != fn.end(); ++i)
        {
            stat(i->c_str(), &st);
            total += st.st_size;
        }
        
        tasks.push_back(rec(p, total, key));
        
        if(tasks.size() >= processlimit)
            laststatus = wait(NULL);
    }
    const unsigned size() const
    {
        return tasks.size();
    }
    bool pending(bool force=false)
    {
        for(;;)
        {
            if(laststatus > 0)return true;
        	if(!size())return false;
            if(tasks.size() < processlimit && !force)return false;
            laststatus = wait(NULL);
        }
    }
    unsigned which()
    {
        for(unsigned a=0; a<tasks.size(); ++a)
            if(tasks[a].pid == laststatus)
                return tasks[a].key;
    }
    double fetch()
    {
        for(unsigned a=0; a<tasks.size(); ++a)
            if(tasks[a].pid == laststatus)
            {
                laststatus = -1;
                return finish(a);
            }
    }
};

#include "argh.hh"

similarity_tester tester(1);

static void Killer(int sig)
{
	fprintf(stderr, "\r");
	tester.die();
	signal(sig, SIG_DFL);
	switch(sig)
	{
		case SIGINT: fprintf(stderr, "Interrupted\n"); break;
		case SIGTERM: fprintf(stderr, "Terminated\n"); break;
	}
	raise(sig);
}

int main(int argc, const char *const *argv)
{
	signal(SIGINT, Killer);
	signal(SIGTERM, Killer);
	
    dup2(nul_fd, 0); /* stdin */
    
    bool worst = false;

    vector<string> files;
    
    string buildfn;
    
    ParamHandler Argh;
    
    Argh.AddLong("worst",   'w').SetBool().SetDesc("Least optimal.");
    Argh.AddLong("help",    'h').SetBool().SetDesc("This help.");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Displays version information.");
    Argh.AddLong("build",   'b').SetString().SetDesc("Builds the archive.", "<file>");
    Argh.AddLong("rar",     'r').SetBool().SetDesc("RAR mode. (default)");
    Argh.AddLong("gzip",    'z').SetBool().SetDesc(".tar.gz mode.");
    Argh.AddLong("bzip2",   'y').SetBool().SetDesc(".tar.bz2 mode.");
    Argh.AddLong("make",    'b');
    Argh.AddLong("max-jobs",'j').SetInt(1, 255).SetDesc("Simultanous process limit (default 1).", "<num>");
    
    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'w': worst = Argh.GetBool(); break;
            case 'V': printf("%s\n", VERSION); return 0;
            case 'b': buildfn = Argh.GetString(); break;
            case 'r': op_mode = make_rar; break;
            case 'z': op_mode = make_gzip; break;
            case 'y': op_mode = make_bzip2; break;
            case 'j': tester.set_process_limit(Argh.GetInt()); break;
            case 'h':
                if(Argh.GetBool()) /* Allow --without-help ;) */
                {
                    printf(
                        "Optrar v"VERSION" - Copyright (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/)\n"
                        "\n"
                        "Usage: optrar [<option> [...]] <file> [<...>]\n"
                        "\n"
                        "Builds the optimal list of files for solid archiving.\n"
                        "\n"
                        "Options:\n");
                    Argh.ListOptions();
                    printf(
                        "\n"
                        "These commands were used when testing:\n"
                        "  rar a test.rar -mm -m5 -s main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                        "    # 138211 bytes (this would be default result)\n"
                        "  ./optrar -bbest.rar       main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                        "    # 138027 bytes (this is the optimal case)\n"
                        "  ./optrar -bworst.rar -w   main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                        "    # 139500 bytes (this is the worst case)\n"
                        "\n"
                        "  tar cfz test.tar.gz         main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                        "    # 203168 bytes (this would be default result)\n"
                        "  ./optrar -bbest.tar.gz  -z  main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                        "    # 199922 bytes (this is the optimal case)\n"
                        "  ./optrar -bworst.tar.gz -zw main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                        "    # 201883 bytes (this is the worst case)\n"
                        "\n"
                        "  tar cfy test.tar.bz2         main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                        "    # 160738 bytes (this would be default result)\n"
                        "  ./optrar -bbest.tar.bz2  -y  main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                        "    # 160640 bytes (this is the optimal case)\n"
                        "  ./optrar -bworst.tar.bz2 -yw main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                        "    # 157741 bytes (this is the worst case)\n"
                        "\n"
                        "Notice that all those rar-files contain the same files inside them,\n"
                        "but the ordering of the files affects the size of the result.\n"
                        "\n"
                        "The best/worst -finding apparently happens to work for RAR archives only ;)\n"
                        "\n"
                    );
                    return 0;
                }
                break;
            default:
                files.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    if(!files.size())
    {
        fprintf(stderr, "Error: At least one file must be specified.\n");
        return -1;
    }
    
    const unsigned fc = files.size();
    
    list<string> lista;
    vector<bool> used(fc, false);
    
    fprintf(stderr, "%u files.\nWill do %u test compressions.\n", fc, fc*(fc+1)/2);
    
    for(unsigned depth=0; depth<fc; ++depth)
    {
        double best = 0;
        if(worst)best = 3.4E+38;
        unsigned bestp = 0;
        for(unsigned testno=0; testno<=fc; ++testno)
        {
        	if(testno < fc)
        	{
                if(used[testno])continue;

                list<string> listt = lista;
                listt.push_front(files[testno]);

                tester.add(testno, listt);
            }            
            if(testno == fc){fprintf(stderr, "\rWaiting...        ");fflush(stderr);}
            while(tester.pending(testno == fc))
            {
                unsigned lst = tester.which();
                double score = tester.fetch();

                if((!worst && score > best)
                ||  (worst && score < best))
                {
                    best = score;
                    bestp = lst;
                }
            }
            fprintf(stderr, "\r%.2f%% done", (depth*fc+testno)*100.0/(fc*fc));
            fflush(stderr);
        }

        used[bestp] = true;
        lista.push_front(files[bestp]);
    }
    fprintf(stderr, "%-25s\r", "\r100% done\r"); fflush(stderr);
    
    if(buildfn.size())
    {
    	unlink(buildfn.c_str());
        build_arch(buildfn, lista);
    }
    else
    {
        for(list<string>::const_iterator i=lista.begin(); i!=lista.end(); ++i)
            printf("%s\n", i->c_str());
    }

    fprintf(stderr, "\n");
    
    return 0;
}
